  //*******************************************************//
// Implementation of DBox class
#include <iostream.h>
#include <iomanip.h>
#include <sfftw.h>
#include <math.h>
#include <stdio.h>
#include "curvelet.h"
#include "dmath.h"

#define MEYERCOARSE 0
#define MEYERDETAIL 1
#define MEYERFINE   2

const int APOLDEG=9;

DBox::DBox()
{
  DIM = 3;
//  a0 = 4;  // too small
  a0 = 8;  // 8 is normal; 2*a0 will be a radius of the coarse scale
  nn0 = 8; // 7 parameter defining number of angles for for the first fine scale, k=1
           // in 2D this would correspond to 4*nn0= 28 angles, in 3D 6*nn0^2= 294 angles.
 
  isWinFuncSetup = 0;
  isGridSetup = 0;

  Kend = 1;
  
  K1 = 0;
  N_nu_c = 0;
  N_nu = 0;
  L1 = L2 = 0;
  Kc = 0;
  K1low = 0;

  boxw1 = boxw2 = boxw3 = 0;

  OverSampleFac = 0;
  Nk = 0;
  Ngr = 0;
  NgridsScale=NgridsWinScale=0;   
  WindowFunc = 0; WindowFuncScale = 0;
  Grid[0] = Grid[1] = Grid[2] = 0;
  Pqs_min = Pqs_max=0; 
  flagp2=0; 
  flagp3=0;  
  KK0=KK1=KK2=0;

  NboxTot = NgridsTot = MaxNgridsPerBox = 0;
}

DBox::~DBox()
{
  DMath::del(K1);
  DMath::del(N_nu_c);
  DMath::del(N_nu);
  DMath::del(N_nu_sym);
  DMath::del(Ngr);
  DMath::del(L1);
  DMath::del(L2);
  DMath::del(Kc);
  DMath::del(K1low);

  DMath::del(boxw1);
  DMath::del(boxw2);
  DMath::del(boxw3);

  DMath::del(OverSampleFac);
  DMath::del(Nk);
  DMath::del(WindowFunc); DMath::del(WindowFuncScale);

  DMath::del(Grid[0]);  DMath::del(Grid[1]);  DMath::del(Grid[2]);
  DMath::del(NgridsScale);  DMath::del(NgridsWinScale);
  DMath::del(Pqs_min);  DMath::del(Pqs_max);
  DMath::del(flagp3);  
}

int DBox::SetUp(int N_image, double eps, int flag, int flag1)
{                                 // flag=1  (0)  => calculate window functions and save them into file (not to calculate window functions)
                                   //  flag1=2  => allocate memory for window functions and read them from file
                                   //             1 => just allocate memory
                                   //             0 => do nothing during setup
  double tmp, temp_k = log(0.25*N_image/a0)/log(2.0);
                                   // maximum image radius is N_image/2. Thus largest pseudocenter is N_image/4
  Kend = (int)(temp_k+0.5) + 1; // number of scales (including the coarse csale)
  
  // memory
  FILE *fid;
  K1 = DMath::newD(Kend);
  N_nu_c = DMath::newI(Kend);
  N_nu = DMath::newL(Kend);
  N_nu_sym = DMath::newI(Kend);
  Ngr = DMath::newL(Kend);
  L1 = DMath::newD(Kend);
  L2 = DMath::newD(Kend);
  Kc = DMath::newD(Kend);
  K1low = DMath::newD(Kend);
    
    cout << "Kend=" << Kend << "\n";  cout.flush();

  // set up box geometry
  int k;
  for(k=1;k<Kend;k++)
  {
    K1[k] = a0*pow(2.0,k); // center of scale k

    temp_k = nn0*pow(2.0, (double)(k-1)/2.0);

       N_nu_c[k] = ((int)temp_k)/2*2;   // to make number even
//    N_nu_c[k] = nn0;
    N_nu[k] = 6*N_nu_c[k]*N_nu_c[k]; // total number of boxes on each scale
    N_nu_sym[k] = N_nu_c[k]*N_nu_c[k]/4;    
  }
  K1[0] = 0;
  N_nu_c[0] = 0;
  N_nu[0] = 1;
  N_nu_sym[0] = 1;

// smearing radius 
  double logeps = -log(eps);
  Lamd = PI*PI/(2.0*logeps);
  half_q = (int)(sqrt(logeps/ Lamd)); 


     // memory  for orientation vectors 
  int *accumul = DMath::newI(Kend+1);

  accumul[0] = 0;
  for(k=1;k<=Kend;k++)
    accumul[k] = accumul[k-1] + N_nu[k-1]*3;
  
  int num_box_tot_3 = accumul[Kend];

  boxw1 = new double*[Kend];  boxw1[0] = new double[num_box_tot_3];
  for(k=1;k<Kend;k++)  boxw1[k] = boxw1[0] + accumul[k];

  boxw2 = new double*[Kend];  boxw2[0] = DMath::newD(num_box_tot_3);
  for(k=1;k<Kend;k++)  boxw2[k] = boxw2[0] + accumul[k];

  boxw3 = new double*[Kend];  boxw3[0] = DMath::newD(num_box_tot_3);
  for(k=1;k<Kend;k++)  boxw3[k] = boxw3[0] + accumul[k];
  DMath::del(accumul);

         //cubed sphere for orientation vectors.  
  double box_r;
  for(k=1;k<Kend;k++)
  {
     Cube_sphere_1(N_nu_c[k], DIM, boxw1[k], boxw2[k], boxw3[k], &box_r);

     L2[k] = (2*K1[k])*(2*box_r); // box width
     K1low[k] = sqrt(fmax(K1[k]*K1[k] - L2[k]*L2[k],36.0))/2.0; // radiaus of lower boundary of box
     // K1low[k] = ceil(K1[k]/4.0*cos(6.28318530717959/N_nu_c[k]))*2.0;  // Fredeik's magic formula for lower boundary
     Kc[k] = (2*K1[k]+K1low[k])/2; // box center coordinates
     L1[k] = (2*K1[k]-K1low[k]); // box length
  }
  L1[0] = L2[0] = K1[1]; // box dimension for scale 0, 2*L1[0]+1
  K1low[0] = 0.0;
  Kc[0] = 0.0;



  Pqs_min = DMath::newI(Kend,4);  Pqs_max = DMath::newI(Kend,4); 

  int ii, jj, itmp;    // Distance is doubled (thus we account for oversampling)
  for(k=1;k<Kend;k++) {
      Pqs_min[k][0]=Pqs_min[k][1]=Pqs_min[k][2]=N_image;  Pqs_max[k][0]=Pqs_max[k][1]=Pqs_max[k][2]=-N_image; 
      for(ii=0;ii<N_nu_sym[k];ii++) {
           for(jj=0;jj<3;jj++) {
               itmp=(int)(2.0*K1low[k]*boxw1[k][jj+ii*3]-L2[k]*(boxw2[k][jj+ii*3]+boxw3[k][jj+ii*3])); // x-component  of one box corner
               if(itmp<Pqs_min[k][jj])  Pqs_min[k][jj]=itmp;  if(itmp>Pqs_max[k][jj])  Pqs_max[k][jj]=itmp; 
               itmp=(int)(2.0*K1low[k]*boxw1[k][jj+ii*3]-L2[k]*(boxw2[k][jj+ii*3]-boxw3[k][jj+ii*3])); 
               if(itmp<Pqs_min[k][jj])  Pqs_min[k][jj]=itmp;  if(itmp>Pqs_max[k][jj])  Pqs_max[k][jj]=itmp; 
               itmp=(int)(2.0*K1low[k]*boxw1[k][jj+ii*3]+L2[k]*(boxw2[k][jj+ii*3]+boxw3[k][jj+ii*3])); 
               if(itmp<Pqs_min[k][jj])  Pqs_min[k][jj]=itmp;  if(itmp>Pqs_max[k][jj])  Pqs_max[k][jj]=itmp; 
               itmp=(int)(2.0*K1low[k]*boxw1[k][jj+ii*3]+L2[k]*(boxw2[k][jj+ii*3]-boxw3[k][jj+ii*3])); 
               if(itmp<Pqs_min[k][jj])  Pqs_min[k][jj]=itmp;  if(itmp>Pqs_max[k][jj])  Pqs_max[k][jj]=itmp; 
               itmp=(int)(4.0*K1[k]*boxw1[k][jj+ii*3]-L2[k]*(boxw2[k][jj+ii*3]+boxw3[k][jj+ii*3]));  // x-component  of one box corner
               if(itmp<Pqs_min[k][jj])  Pqs_min[k][jj]=itmp;  if(itmp>Pqs_max[k][jj])  Pqs_max[k][jj]=itmp; 
                itmp=(int)(4.0*K1[k]*boxw1[k][jj+ii*3]-L2[k]*(boxw2[k][jj+ii*3]-boxw3[k][jj+ii*3])); 
               if(itmp<Pqs_min[k][jj])  Pqs_min[k][jj]=itmp;  if(itmp>Pqs_max[k][jj])  Pqs_max[k][jj]=itmp; 
               itmp=(int)(4.0*K1[k]*boxw1[k][jj+ii*3]+L2[k]*(boxw2[k][jj+ii*3]+boxw3[k][jj+ii*3])); 
               if(itmp<Pqs_min[k][jj])  Pqs_min[k][jj]=itmp;  if(itmp>Pqs_max[k][jj])  Pqs_max[k][jj]=itmp; 
                itmp=(int)(4.0*K1[k]*boxw1[k][jj+ii*3]+L2[k]*(boxw2[k][jj+ii*3]-boxw3[k][jj+ii*3])); 
               if(itmp<Pqs_min[k][jj])  Pqs_min[k][jj]=itmp;  if(itmp>Pqs_max[k][jj])  Pqs_max[k][jj]=itmp; 
           } // for(jj=0...
      } // for(ii=0...
      for(jj=0;jj<3;jj++) { Pqs_min[k][jj]=max(Pqs_min[k][jj]-half_q, -N_image); Pqs_max[k][jj]=min(Pqs_max[k][jj]+half_q+1, N_image); }
      Pqs_min[k][3]=min(Pqs_min[k][0],Pqs_min[k][2]); Pqs_max[k][3]=min(N_image,Pqs_max[k][2]);
      Pqs_min[k][3]=min(Pqs_min[k][0],Pqs_min[k][2]); Pqs_max[k][3]=min(N_image,Pqs_max[k][2]);
  } // for(kk=0...
  Pqs_min[0][0]=Pqs_min[0][1]=Pqs_min[0][2]=Pqs_min[0][3]=-(int)(2.0*L2[0])-half_q;
  Pqs_max[0][0]=Pqs_max[0][1]=Pqs_max[0][2]=Pqs_max[0][3]=(int)(2.0*L2[0])+half_q+1;
               // make index span form 0 to 2*N_image
//   for(k=0;k<Kend;k++)  for(jj=0;jj<4;jj++) {Nqs_min[k][jj]+=N_image; Nqs_max[k][jj]+=N_image; }


  printf("L1= ");  for(k=0;k<Kend;k++)  printf("%f, ", L1[k]);   printf("...\n");
  printf("K1low= ");  for(k=0;k<Kend;k++)  printf("%f, ", K1low[k]);   printf("...\n");  
//  printf("N_nu= ");  for(k=0;k<Kend;k++)  printf("%f, ", N_nu[k]);   printf("...\n");  
 
  setup_Nk(eps); // number of grids


 // Calculate total number of grids (boxes) and grid points

  double dMbconv=8.0/1024.0/1024.0, fMbconv=4.0/1024.0/1024.0;
  long  mem_gr, mem_win, mem_dat_scale, mem_dat_quad, mem_dat_sect, mem;
  NgridsScale=DMath::newL(Kend); NgridsWinScale=DMath::newL(Kend);   
  flagp3=DMath::newI(Kend); 

  NboxTot = 0; NgridsTot = 0;  NgridsWinTot = 0;  MaxNgridsPerBox = 0;
  for(k=0;k<Kend;k++)
  {
     NboxTot += N_nu[k];
  
     Ngr[k] = Nk[0][k]*Nk[1][k]*Nk[2][k];
     NgridsScale[k] = Ngr[k]*N_nu[k]; NgridsTot += NgridsScale[k];
     NgridsWinScale[k] = Ngr[k]*N_nu_sym[k]; NgridsWinTot += NgridsWinScale[k];
     if(MaxNgridsPerBox<Ngr[k]) MaxNgridsPerBox = Ngr[k];    
     
     mem_gr=(long)((double)Ngr[k]*3.0*fMbconv); 
     mem_win=(long)((double)NgridsWinScale[k]*fMbconv); 
     mem_dat_scale=(long)(2.0*fMbconv*pow(2*2*(2*K1[k]),3));
     tmp=(double)(Pqs_max[k][3]-Pqs_min[k][3]);
     mem_dat_quad=(long)(tmp*tmp*tmp*2.0*fMbconv); 
     mem_dat_sect=(long)((Pqs_max[k][0]-Pqs_min[k][0])*(Pqs_max[k][1]-Pqs_min[k][1])*(Pqs_max[k][2]-Pqs_min[k][2])*2.0*fMbconv); 
     mem=mem_gr+mem_win+MAX_REDUCE_BUF_SIZE;
     mem=mem_gr/3.0*4.0+mem_win+MAX_REDUCE_BUF_SIZE; // delete me !!!!

     if((mem+mem_dat_scale+150) < MEMORY_PER_NODE) KK0=k; 
     if((mem+mem_dat_quad+150) < MEMORY_PER_NODE) KK1=k;  
     if((mem+mem_dat_quad/8+150) < MEMORY_PER_NODE) KK2=k;
     cout << "scale=" << k << ", N_nu[k]="<< N_nu[k]  << ", (N1,N2)=("<< Nk[0][k] << "," << Nk[1][k] << "), mem(grbox+win): " << (mem_gr+mem_win) << " Mb.  CompCost [sc,quad,sect]: [" << (int)(NgridsScale[k]*0.0001)  << ", " << (int)(3*NgridsWinScale[k]*0.0001) << ", " << (int)(NgridsWinScale[k]*0.0001) << "];\n";  cout.flush();
     cout <<  "      mem(data quad): "<< mem_dat_quad << " Mb, mem(data qsect): "<< mem_dat_sect << " Mb....." << Pqs_min[k][3] << ","<<  Pqs_max[k][3] << "; sect y range: "<< Pqs_min[k][1] << ","<<  Pqs_max[k][1] << "; sect z range: "<< Pqs_min[k][2] << ","<<  Pqs_max[k][2]<< ".\n";  cout.flush();
  }
  if(flagp2==1) printf("\n <<<< Problem >>>>>>>: be careful: you may have not enough memory to store boxgrid and window functions. \n"); 
  if(KK0<(Kend-1)) printf("\n <<<< Problem >>>>>>>: Program can not deal with this size of image yet. \n"); 

  printf("\n KK0: %d, KK1: %d, KK2: %d \n", KK0, KK1, KK2); 
  printf("\n total number of coefficients: %d, memory: %d Mb \n", (int)NgridsTot, (int)((double)(NgridsTot)*2.0*fMbconv)); 
 
  int tmmp;
  if(flag) {   // writing configuration and precomputing window functions  
                 //  write configuration into file 
        fid = fopen("./CCoeff/conf_info","wb");
        fwrite(&Kend, sizeof(int), 1, fid);  
        for(k=0;k<Kend;k++) { tmmp=(int)(N_nu[k]); fwrite(&tmmp, sizeof(int), 1, fid); }
        for(k=0;k<Kend;k++) fwrite(&Nk[0][k], sizeof(int), 1, fid);  for(k=0;k<Kend;k++) fwrite(&Nk[1][k], sizeof(int), 1, fid); 
        fwrite(&L1[0], sizeof(double), Kend, fid);  fwrite(&L2[0], sizeof(double), Kend, fid);
        fwrite(&K1low[0], sizeof(double), Kend, fid);  fwrite(&K1[0], sizeof(double), Kend, fid);
        fwrite(&boxw1[0][0], sizeof(double), num_box_tot_3, fid); fwrite(&boxw2[0][0], sizeof(double), num_box_tot_3, fid); fwrite(&boxw3[0][0], sizeof(double), num_box_tot_3, fid);
        fclose (fid);
              // calc window functions
    setup_window_func(); 
  }

  if(flag1>0) {  
      WindowFunc = allocD(Kend, N_nu_sym, Nk); 
     if(flag1==2) {  
                   // read window functions from file
        fid = fopen("./CCoeff/Wind_func_all.dat","rb");
        fread(&WindowFunc[0][0][0],sizeof(float),NgridsWinTot,fid);
        fclose (fid);  }
  }

   return 0;
}




// *************************************************************
// other functions implementation
// *************************************************************

    // oversampling and grid size on boxes

void DBox::setup_Nk(double eps) 
{
  OverSampleFac = DMath::newD(3, Kend);
  Nk = DMath::newI(3, Kend+1); // number of grid points
  
  double L[2];
  int N[3];
  int k;
  double os_fac[3];

  for(k=0;k<Kend;k++)
    {
      L[0] = L1[k];  L[1] = L2[k];
      get_os(k, L, eps, N, os_fac);

      OverSampleFac[0][k] = os_fac[0];   OverSampleFac[1][k] = os_fac[1];  OverSampleFac[2][k] = os_fac[2];
      Nk[0][k] = N[0];  Nk[1][k] = N[1];  Nk[2][k] = N[2];
    }
}





    //  estimate oversampling for boxes at each scale
void DBox::get_os(int scale, double *L, double eps, int *N, double *over_sample)
{
  int j,k,n, type;
  float *y1, *y2, *y3; 
  double *w;
  bool check;
  int M[3];
  int LL1,LL2;
  double a0, a1, ar, ai;
  double nu1, nu2, step;
  fftw_complex *in, *out;
  fftw_plan p;


  if(scale != 0) {
    LL1=(int)((L[0]+1)/2)*2; LL2=(int)((L[1]+1)/2)*2;
    if(scale < (Kend-1)) type=MEYERDETAIL; else type=MEYERFINE;

    // radial direction
    n=6*LL1; step=L[0]/(double)n;
    M[0]=n; M[1]=1; M[2]=1;
    y1=(float*)malloc(n*sizeof(float)); y2=(float*)malloc(n*sizeof(float)); y3=(float*)malloc(n*sizeof(float));
    w=(double*)malloc(n*sizeof(double));
    in = (fftw_complex *)malloc(n*sizeof(fftw_complex));   out = (fftw_complex *)malloc(n*sizeof(fftw_complex));
    p = fftw_create_plan(n, FFTW_BACKWARD, FFTW_ESTIMATE);
 
 
    for (j=0;j<n;j++){ y1[j]=K1low[scale]+step*j; y2[j]=0; y3[j]=0; }
    chihat_rad(y1,y2,y3,w,M,K1[scale],type);
    for (j=0;j<n;j++)  {in[j].re=w[j]; in[j].im=0; }

    fftw_one(p, in, out);
    a0=0;
    for (j=0;j<n/2;j++)  {w[j]=out[j].re*out[j].re+out[j].im*out[j].im;  if(a0<w[j]) a0=w[j]; }
    j=0;check=true;
    while(check){ j++; check=((w[j]/a0>eps)&&(j<n/2)); }
    N[0]=LL1+2*j;
    nu1=((double)N[0])/L[0];


    printf("Oversampling: k=%d, overs_r=%f, j_r=%d, L1=%f, LL1=%d, ",scale, nu1,j,L[0],LL1);

    free(y1);free(y2);free(y3);free(w);
    fftw_destroy_plan(p);
    free(in); free(out);

        // orthogonal direction
    n=6*LL2; step=L[1]/(double)n;
    M[0]=1;M[1]=n;M[2]=1;
    y1=(float*)malloc(n*sizeof(float)); y2=(float*)malloc(n*sizeof(float)); y3=(float*)malloc(n*sizeof(float));
    w=(double*)malloc(n*sizeof(double));
    in = (fftw_complex *)malloc(n*sizeof(fftw_complex));   out = (fftw_complex *)malloc(n*sizeof(fftw_complex));
    p = fftw_create_plan(n, FFTW_BACKWARD, FFTW_ESTIMATE);
      
    for(j=0;j<n;j++){ y1[j]=K1[scale];y2[j]=-L[1]/2.0 + step*j; y3[j]=0;};
    chihat_ang(y1,y2,y3,w,M,L,K1[scale]);
    for (j=0;j<n;j++)  {in[j].re=w[j]; in[j].im=0; }
     
    fftw_one(p, in, out);
    a0=0;
    for (j=0;j<n/2;j++)  {w[j]=out[j].re*out[j].re+out[j].im*out[j].im;  if(a0<w[j]) a0=w[j]; }
    j=0;check=true;
    while(check){ j++; check=((w[j]/a0>eps)&&(j<n/2)); }
    N[1]=LL2+2*j;
    nu2=((double)N[1])/L[1];


    // a0=0; for(k=0;k<n;k++) a0+=w[k];

    printf("overs_a=%f, j_a=%d, L2=%f, LL2=%d...\n",nu2,j,L[1],LL2);
    free(y1);free(y2);free(y3);free(w);
    fftw_destroy_plan(p);
    free(in); free(out);

    N[2]=N[1];
    over_sample[0] = nu1;
    over_sample[1] = over_sample[2] = nu2;
  }

  else {
  // scale 0

     LL1=(int)(L[0])*2; LL2=(int)(L[1])*2;
     n=6*LL1; step=2.0*L1[0]/((double)n);
  M[0]=n; M[1]=1; M[2]=1;
  y1=(float*)malloc(n*sizeof(float)); y2=(float*)malloc(n*sizeof(float)); y3=(float*)malloc(n*sizeof(float));
  w=(double*)malloc(n*sizeof(double));
  in = (fftw_complex *)malloc(n*sizeof(fftw_complex));   out = (fftw_complex *)malloc(n*sizeof(fftw_complex));
  p = fftw_create_plan(n, FFTW_BACKWARD, FFTW_ESTIMATE);

  for (j=0;j<n;j++){ y1[j]=-L1[0]+step*j; y2[j]=0; y3[j]=0; }
  chihat_rad(y1,y2,y3,w,M,L1[0],MEYERCOARSE);  
  for (j=0;j<n;j++)  {in[j].re=w[j]; in[j].im=0; }
     
  fftw_one(p, in, out);
  a0=0.0;
  for (j=0;j<n/2;j++)  {w[j]=out[j].re*out[j].re+out[j].im*out[j].im;  if(a0<w[j]) a0=w[j]; }
  j=0;check=true;
  while(check){ j++; check=((w[j]/a0>eps)&&(j<n/2)); }
  N[0]=LL1+2*j;
  nu1=((double)N[0])/(2.0*L[0]);

  //nu1=((double)(2*j))/((double)(LL1+1));
   printf("Oversampling: k=%d, overs=%f, j=%d, L1=%f, LL1=%d... \n ",scale, nu1,j,2*L[0],LL1);
  free(y1);free(y2);free(y3);free(w);

  N[1] = N[2] = N[0];
  over_sample[0] = over_sample[1] = over_sample[2] = nu1;
 }  

  return;
}








// Pre-calculate window function values on all the grid poins

void DBox::setup_window_func()
{
  if(isWinFuncSetup) return;

   FILE *fid;
  int k;
  double L[2];
  int N[3]; 

  // window functions
  long ngrids_max = 0, N_nu_max = 0;
  long j, ngrids_local;
  int j1, j2, j3;
  int ix_angle;
  double step_x1, step_x2, step_x3, K2low;
  double covers_norm;
 
  NgridsWinTot=0;
  for(k=0;k<Kend;k++) {    
    if(N_nu_max<N_nu[k]) N_nu_max = N_nu[k];
    ngrids_local = Nk[0][k]*Nk[1][k]*Nk[2][k];
    NgridsWinTot+= ngrids_local*N_nu_sym[k];
    if(ngrids_max<ngrids_local) ngrids_max = ngrids_local;
  } 

  float *x1,*x2,*x3;
  x1=(float*) malloc(ngrids_max*sizeof(float));
  x2=(float*) malloc(ngrids_max*sizeof(float));
  x3=(float*) malloc(ngrids_max*sizeof(float));

  double *p1, *p2, *p3;
 
    // scale 0
    float **wf; wf = DMath::newF(N_nu_sym[0], Ngr[0]);  // memory

    CalcGrid(0, x1, x2, x3);
    L[0] = L1[1]; L[1] = L2[1];  // not used 
    p1 = &boxw1[k][0];   p2 = &boxw2[k][0];    p3 = &boxw3[k][0];  // not used 
    
    N[0]=Nk[0][0]; N[1]=Nk[1][0]; N[2]=Nk[2][0]; 
    chihat_f(x1, x2, x3, p1, p2, p3, &wf[0][0], N, L, K1[1], N_nu[1], 0, 1); // actually used input: x1, x2, x3, window_func, N, K1[1], 0

   fid = fopen("./CCoeff/Wind_func_all.dat","wb");  fwrite(&wf[0][0],sizeof(float),(Nk[0][0]*Nk[1][0]*Nk[2][0]*N_nu_sym[0]),fid);  fclose (fid);
   DMath::del(wf);

    // scale 1 to Kend-1
  for(k=1; k<Kend; k++) 
 {
    printf("...generating window functions for scale: %d...\n",k);
    float **wf; wf = DMath::newF(N_nu_sym[k], Ngr[k]);  // memory

    CalcGrid(k, x1, x2, x3);
    L[0] = L1[k]; L[1] = L2[k];
    N[0]=Nk[0][k]; N[1]=Nk[1][k]; N[2]=Nk[2][k];   
    p1 = &boxw1[k][0];   p2 = &boxw2[k][0];   p3 = &boxw3[k][0];

    for(ix_angle=0; ix_angle<N_nu_sym[k]; ix_angle++) 
    {
        printf("...angle: %d...\n",ix_angle);
        if(L2[k] < (2*K1low[k]))
            { chihat_sm_f(x1, x2, x3, p1, p2, p3, &wf[ix_angle][0], N, L, K1[k], N_nu[k], k,ix_angle); }
        else
            { chihat_f(x1, x2, x3, p1, p2, p3, &wf[ix_angle][0], N, L, K1[k], N_nu[k], k, ix_angle); }
     } // for(ix_angle=0...
  fid = fopen("./CCoeff/Wind_func_all.dat","ab");  fwrite(&wf[0][0],sizeof(float),(Nk[0][k]*Nk[1][k]*Nk[2][k]*N_nu_sym[k]),fid);  fclose (fid);
  DMath::del(wf);
  } // for(k=1...
   
  // clean up
   free(x1); free(x2); free(x3);

  isWinFuncSetup = 1;
}


int DBox::GetWindowFunctScale(int k1)
{       
  FILE *fid;
  long k, stride = 0;
  for(k=0;k<k1;k++)  stride += NgridsWinScale[k];

   fid = fopen("./CCoeff/Wind_func_all.dat","rb");
   fseek (fid , stride*sizeof(float), SEEK_SET);
   fread(&WindowFuncScale[0][0],sizeof(float),NgridsWinScale[k1],fid);
   fclose (fid);  

   return 0;
}




// *********************************************************
// working with individual boxes

// my function calculating grid for scale k aligned along axis x1
int DBox::CalcGrid(int k, float *yx1, float *yx2, float *yx3)
{
  int j1, j2, j3;
  long ngrids = Nk[0][k]*Nk[1][k]*Nk[2][k];

  double step_x1, step_x2, step_x3, K2low;
  double L[2], ThetInv[3][3];
  int N[3];

  N[0] = Nk[0][k]; N[1] = Nk[1][k]; N[2] = Nk[2][k];

  // scale 0
   if(k==0)
  {
     step_x1 = 2.0*L1[0]/((double)N[0]); 
  
    for(j1=0;j1<N[0];j1++) //Compute the grid points, but normalized to be aligned along the x1-direction.
          for(j2=0;j2<N[1];j2++)
	for(j3=0;j3<N[2];j3++){
                       yx1[j3+N[2]*(j2+N[1]*j1)]= -L1[0] + j1*step_x1;
                       yx2[j3+N[2]*(j2+N[1]*j1)]= -L1[0] + j2*step_x1;
                       yx3[j3+N[2]*(j2+N[1]*j1)]= -L1[0] + j3*step_x1;
                  }
  } // end if

  // scale 1 to Kend-1
  else
  {
    L[0] = L1[k]; L[1] = L2[k];
    K2low = L2[k]*0.5;

    step_x1 = L[0]/((double)N[0]);  step_x2 = L[1]/((double)N[1]);  step_x3 = L[1]/((double)N[2]);

    for(j1=0;j1<N[0];j1++)//Compute the grid points, but normalized to be aligned along the x1-direction.
          for(j2=0;j2<N[1];j2++)
                for(j3=0;j3<N[2];j3++){
                    yx1[j3+N[2]*(j2+N[1]*j1)]= K1low[k] + j1*step_x1;
                    yx2[j3+N[2]*(j2+N[1]*j1)]=-K2low + j2*step_x2;
                    yx3[j3+N[2]*(j2+N[1]*j1)]=-K2low + j3*step_x3;
                }
  } // end else

  return 0;
}

// my function calculating grrid for the box k, ix_angle 
int DBox::CalcGrid(int k, int ix_angle, float *yx1, float *yx2, float *yx3)
{
  int j1, j2, j3;
  long ind, ngrids = Nk[0][k]*Nk[1][k]*Nk[2][k];

  float l1,l2,l3;

  double step_x1, step_x2, step_x3, K2low;
  double L[2], ThetInv[3][3];
  int N[3];

  N[0] = Nk[0][k]; N[1] = Nk[1][k]; N[2] = Nk[2][k];

  // scale 0
   if(k==0)
  {
     step_x1 = 2.0*L1[0]/((double)N[0]); 
  
    for(j1=0;j1<N[0];j1++) //Compute the grid points, but normalized to be aligned along the x1-direction.
       for(j2=0;j2<N[1];j2++)
            for(j3=0;j3<N[2];j3++){
          yx1[j3+N[2]*(j2+N[1]*j1)]=-L1[0] + j1*step_x1;
          yx2[j3+N[2]*(j2+N[1]*j1)]=-L1[0] + j2*step_x1;
          yx3[j3+N[2]*(j2+N[1]*j1)]=-L1[0] + j3*step_x1;
        }
  } // end if

  // scale 1 to Kend-1
  else
  {
    L[0] = L1[k]; L[1] = L2[k];
    K2low = L2[k]*0.5;

    step_x1 = L[0]/((double)N[0]); step_x2 = L[1]/((double)N[1]); step_x3 = L[1]/((double)N[2]);
      
  // rotate grids to the correct angle

  ThetInv[0][0] = boxw1[k][ix_angle*3]; ThetInv[1][0] = boxw1[k][ix_angle*3+1]; ThetInv[2][0] = boxw1[k][ix_angle*3+2];
  ThetInv[0][1] = boxw2[k][ix_angle*3]; ThetInv[1][1] = boxw2[k][ix_angle*3+1]; ThetInv[2][1] = boxw2[k][ix_angle*3+2];
  ThetInv[0][2] = boxw3[k][ix_angle*3]; ThetInv[1][2] = boxw3[k][ix_angle*3+1]; ThetInv[2][2] = boxw3[k][ix_angle*3+2];

    for(j1=0;j1<N[0];j1++)//Compute the grid points, but normalized to be aligned along the x1-direction.
        for(j2=0;j2<N[1];j2++)
           for(j3=0;j3<N[2];j3++){
               ind=j3+N[2]*(j2+N[1]*j1);
               l1= K1low[k] + j1*step_x1;  l2=-K2low + j2*step_x2;  l3=-K2low + j3*step_x3;
               yx1[ind] = ThetInv[0][0]*l1 + ThetInv[0][1]*l2 + ThetInv[0][2]*l3;
               yx2[ind] = ThetInv[1][0]*l1 + ThetInv[1][1]*l2 + ThetInv[1][2]*l3;
               yx3[ind] = ThetInv[2][0]*l1 + ThetInv[2][1]*l2 + ThetInv[2][2]*l3;
           }

  } // end else

  return 0;
}


   // my function calculating window function for the box k, ix_angle 

//****************************************************************//


float*** DBox::allocD(int nscales, int *nangles, int **ngrids)
{
  float ***temp=0;
  int k, i;
  long accumul;

  long NumAnglesTot = 0;
  for(k=0;k<nscales;k++) NumAnglesTot += nangles[k];

  temp = new float**[nscales];
  temp[0] = new float*[NumAnglesTot];
  accumul = 0;
  for(k=1;k<nscales;k++){
    accumul += nangles[k-1];
    temp[k] = temp[0] + accumul;
  }

  long num_grids_tot = 0;
  for(k=0;k<nscales;k++)
    num_grids_tot += nangles[k]*ngrids[0][k]*ngrids[1][k]*ngrids[2][k];
  temp[0][0] = new float[num_grids_tot];

  accumul = 0;
  for(k=0;k<nscales;k++)
    for(i=0;i<nangles[k];i++) {
      temp[k][i] = temp[0][0] + accumul;
      accumul += ngrids[0][k]*ngrids[1][k]*ngrids[2][k];
    }

  return temp;  
}




//************************************************************************//

int DBox::Cube_sphere_1(int NN, int dim, double *w1, double *w2, double *w3, double *r_box)
{
// INPUT:
// !!!! NN is an odd number = number of grid points in one direction on one side
// dim  - dimension of space (3 in our case)
//OUTPUT
// r_box - maximal distance between points on unit sphere
// w1 - set of radial vectors w1[6*NN*NN][w11,w12,w13]
// w2 - set of orthogonal vectors w2[6*NN*NN][w11,w12,w13]

const int NNt=102;
int ii, jj, kk, ind, Ncent, Nh; // Nh - half-grid (different for odd NN) 
double ttn, rr;
double str, tmp, dN_half, normt;
double x_f[NNt], y_f0[NNt], y_f1[NNt], l_arc[NNt];
double *x_sph, *y_sph, *z_sph, *l_arc_reg, *loc_vc, *loc_vcc, *dloc_vc, *x_new, *y_new, *x_cube, *y_cube, *x, *y, *z, *x1, *y1, *z1;

   dN_half = 1.0/((double)NN);
   if(NN%2) {  // odd
         Nh=(NN-1)/2; str=2.0*dN_half; } 
   else {           // even
         Nh=NN/2; str=dN_half; } 

    for(ii=0; ii<NNt; ii++) {  x_f[ii]=ii*.01;   y_f0[ii]=0;   y_f1[ii]=1;  }

//   =============== STEP 1 = make central line even ======================
//     % we take the graphic of horisontal distances between points vvmy1 (the graphic will be l_arc)
//     % note that it defines mapping of distance on cube side x_f to distance on sphere l_arc
//     % we take regular grid l_arc_reg (in l_arc axis) and project it back on (uneven) grid loc_vc for central line on cube 

     // projection of line to the sphere and arclength function (l_arc)
     x_sph = (double*) malloc(sizeof(double)*NNt);  y_sph = (double*) malloc(sizeof(double)*NNt);  z_sph = (double*) malloc(sizeof(double)*NNt);
     for(ii=0; ii<NNt; ii++) {
         tmp=1/sqrt(x_f[ii]*x_f[ii] + y_f0[ii]*y_f0[ii] + 1);    
         x_sph[ii]=x_f[ii]*tmp;  
         y_sph[ii]=y_f0[ii]*tmp;
         z_sph[ii]=tmp;
     }
     for(ii=0,tmp=0; ii<(NNt-1); ii++) {
         tmp=tmp+sqrt((x_sph[ii+1]-x_sph[ii])*(x_sph[ii+1]-x_sph[ii]) + (y_sph[ii+1]-y_sph[ii])*(y_sph[ii+1]-y_sph[ii]) + (z_sph[ii+1]-z_sph[ii])*(z_sph[ii+1]-z_sph[ii]));    
         l_arc[ii+1]=tmp;
     };
     l_arc[0]=0;
     normt=1/l_arc[NNt-1];
     for(ii=0; ii<NNt; ii++)   l_arc[ii]=l_arc[ii]*normt;

// we want to have regular grid in arclength

    l_arc_reg = (double*) malloc(sizeof(double)*(Nh+2));
    for(ii=0; ii<Nh; ii++) {
        l_arc_reg[ii+1]=str+ii*(2*dN_half);
    };
    l_arc_reg[0]=0.0;  l_arc_reg[Nh+1]=1.0;


// interpolation of regular grid in l_arc (l_arc_reg) into grid in x direction (loc_vc)

    loc_vc = (double*) malloc(sizeof(double)*(Nh+2));
    for(ii=1; ii<(Nh+1); ii++) {
        jj=0;  do { jj=jj+1; } while (l_arc[jj] < l_arc_reg[ii]);
        loc_vc[ii] = x_f[jj-1] + (x_f[jj]-x_f[jj-1])*(l_arc_reg[ii]-l_arc[jj-1])/(l_arc[jj]-l_arc[jj-1]);
    };
    loc_vc[0]=0.0; loc_vc[Nh+1]=1.0;    


//   =============== STEP 2 = the same stuff but make boundary line even ======================

     // projection of line to the sphere and arclength function (l_arc)
     for(ii=0; ii<NNt; ii++) {
        tmp=1/sqrt(x_f[ii]*x_f[ii] + y_f1[ii]*y_f1[ii] + 1);    
        x_sph[ii]=x_f[ii]*tmp;
        y_sph[ii]=y_f0[ii]*tmp;
        z_sph[ii]=tmp;
     };
     for(ii=0,tmp=0; ii<(NNt-1); ii++) {
        tmp=tmp+sqrt((x_sph[ii+1]-x_sph[ii])*(x_sph[ii+1]-x_sph[ii]) + (y_sph[ii+1]-y_sph[ii])*(y_sph[ii+1]-y_sph[ii]) + (z_sph[ii+1]-z_sph[ii])*(z_sph[ii+1]-z_sph[ii]));    
        l_arc[ii+1]=tmp;
     };
     l_arc[0]=0;
     normt=1/l_arc[NNt-1];
     for(ii=0; ii<NNt; ii++) {
        l_arc[ii]=l_arc[ii]*normt;
     }; 

// interpolation of regular grid in l_arc (l_arc_reg) into grid in x direction (loc_vc)

    loc_vcc = (double*) malloc(sizeof(double)*(Nh+2));
    dloc_vc = (double*) malloc(sizeof(double)*(Nh+2));
    for(ii=1; ii<(Nh+1); ii++) {
        jj=0;   do { jj=jj+1; } while (l_arc[jj] < l_arc_reg[ii]);
        loc_vcc[ii] = x_f[jj-1] + (x_f[jj]-x_f[jj-1])*(l_arc_reg[ii]-l_arc[jj-1])/(l_arc[jj]-l_arc[jj-1]);
        dloc_vc[ii]=loc_vcc[ii]-loc_vc[ii];
    };
    loc_vcc[0]=0;  loc_vcc[Nh+1]=1; 
    dloc_vc[0]=0;  dloc_vc[Nh+1]=0;    


//   =============== STEP 3 = construct grid for the quater-sector  ======================

    x_new = (double*) malloc(sizeof(double)*Nh*Nh);
    y_new = (double*) malloc(sizeof(double)*Nh*Nh);
    for(jj=0; jj<(Nh*Nh); jj++) { x_new[jj]=0; y_new[jj]=0; }
    
    double tmp1, tmp2;
    double xtx[21], testx[21];
    for(ii=0; ii<20; ii++) {
        xtx[ii]=ii*0.05;
    };
    xtx[20]=1;

    for(ii=0; ii<Nh; ii++) {
        for(kk=0; kk<Nh; kk++) {
            ind=0;
            for(jj=1; jj<21; jj++) {
               tmp1=dloc_vc[kk+1]*xtx[jj]*xtx[jj]+loc_vc[kk+1];
               tmp2=dloc_vc[ii+1]*tmp1*tmp1+loc_vc[ii+1];
               testx[jj]=tmp2-xtx[jj];
            }
            ind=0;
            do { ind=ind+1; } while (testx[ind] > 0);
            tmp2=xtx[ind-1]+ (xtx[ind]-xtx[ind-1])*fabs((testx[ind-1]-0)/(testx[ind]-testx[ind-1]));
            tmp1=dloc_vc[kk+1]*tmp2*tmp2+loc_vc[kk+1];
            x_new[ii*Nh+kk]= tmp1; 
            y_new[kk*Nh+ii]= tmp1;   
        } // for(kk=0...
    } // for(ii=0...

//   =============== STEP 4 = construct grid for the quater-sector  ======================

    x_cube = (double*) malloc(sizeof(double)*(Nh*Nh));
    y_cube = (double*) malloc(sizeof(double)*(Nh*Nh));

    for(ii=0; ii<Nh*Nh; ii++) {  x_cube[ii]=x_new[ii]; y_cube[ii]=y_new[ii];  }

    double xt, yt, zt, x1t, y1t, z1t, x2t, y2t, z2t, bisx, bisy, bisz;
    double vx, vy, vz;
    int Nquater=Nh*Nh;
    int indd1, indd2, indd3;
    
   for(kk=0; kk<Nquater; kk++) {
       tmp=1/sqrt(x_cube[kk]*x_cube[kk] + y_cube[kk]*y_cube[kk] + 1.0);
       xt=x_cube[kk]*tmp; yt=y_cube[kk]*tmp;  zt=tmp;
       x1t=0;  tmp= 1/sqrt(yt*yt+zt*zt); y1t=fabs(zt)*tmp;  z1t=-yt*zt/fabs(zt)*tmp;
       y2t=0;  tmp= 1/sqrt(xt*xt+zt*zt); x2t=fabs(zt)*tmp;  z2t=-xt*zt/fabs(zt)*tmp;
       bisx=x1t+x2t;  bisy=y1t+y2t; bisz=z1t+z2t; tmp= 1/sqrt(bisx*bisx+bisy*bisy+bisz*bisz); bisx*=tmp;  bisy*=tmp; bisz*=tmp;
       vx = yt*bisz - zt*bisy;  vy = zt*bisx - xt*bisz;  vz = xt*bisy - yt*bisx; 
       x1t=vx+bisx; y1t=vy+bisy; z1t=vz+bisz; tmp= 1/sqrt(x1t*x1t+y1t*y1t+z1t*z1t);  x1t*=tmp;  y1t*=tmp;  z1t*=tmp;
        
       w1[3*kk]=xt; w1[1+3*kk]=yt; w1[2+3*kk]=zt;
       w2[3*kk]=x1t; w2[1+3*kk]=y1t; w2[2+3*kk]=z1t;

       w3[3*kk] = -w1[1+3*kk]*w2[2+3*kk] + w1[2+3*kk]*w2[1+3*kk];
       w3[1+3*kk] = -w1[2+3*kk]*w2[3*kk] + w1[3*kk]*w2[2+3*kk];
       w3[2+3*kk] = -w1[3*kk]*w2[1+3*kk] + w1[1+3*kk]*w2[3*kk];
  }

    for(kk=0; kk<Nquater; kk++) {
      indd1=(Nquater+kk); indd2=(2*Nquater+kk);
      w1[3*indd1]=w1[2+3*kk]; w1[1+3*indd1]=w1[3*kk]; w1[2+3*indd1]=w1[1+3*kk];        
      w1[3*indd2]=w1[1+3*kk]; w1[1+3*indd2]=w1[2+3*kk]; w1[2+3*indd2]=w1[3*kk];        
      w2[3*indd1]=w2[2+3*kk]; w2[1+3*indd1]=w2[3*kk]; w2[2+3*indd1]=w2[1+3*kk];        
      w2[3*indd2]=w2[1+3*kk]; w2[1+3*indd2]=w2[2+3*kk]; w2[2+3*indd2]=w2[3*kk];     
      w3[3*indd1]=w3[2+3*kk]; w3[1+3*indd1]=w3[3*kk]; w3[2+3*indd1]=w3[1+3*kk];        
      w3[3*indd2]=w3[1+3*kk]; w3[1+3*indd2]=w3[2+3*kk]; w3[2+3*indd2]=w3[3*kk];     
  }
   for(kk=0; kk<3*Nquater; kk++) {
      indd1=3*Nquater+kk;
      w1[3*indd1]=w1[3*kk]; w1[1+3*indd1]=-w1[1+3*kk]; w1[2+3*indd1]=w1[2+3*kk];        
      w2[3*indd1]=w2[3*kk]; w2[1+3*indd1]=-w2[1+3*kk]; w2[2+3*indd1]=w2[2+3*kk];
      w3[3*indd1]=w3[3*kk]; w3[1+3*indd1]=-w3[1+3*kk]; w3[2+3*indd1]=w3[2+3*kk];
   }
   for(kk=0; kk<6*Nquater; kk++) {
      indd1=6*Nquater+kk;
      w1[3*indd1]=-w1[3*kk]; w1[1+3*indd1]=w1[1+3*kk]; w1[2+3*indd1]=w1[2+3*kk];        
      w2[3*indd1]=-w2[3*kk]; w2[1+3*indd1]=w2[1+3*kk]; w2[2+3*indd1]=w2[2+3*kk];
      w3[3*indd1]=-w3[3*kk]; w3[1+3*indd1]=w3[1+3*kk]; w3[2+3*indd1]=w3[2+3*kk];
   }
    for(kk=0; kk<(3*NN*NN); kk++) {
      indd1=3*(NN*NN)+kk;
      w1[3*indd1]=-w1[3*kk]; w1[1+3*indd1]=-w1[1+3*kk]; w1[2+3*indd1]=-w1[2+3*kk];        
      w2[3*indd1]=-w2[3*kk]; w2[1+3*indd1]=-w2[1+3*kk]; w2[2+3*indd1]=-w2[2+3*kk];        
      w3[3*indd1]=-w3[3*kk]; w3[1+3*indd1]=-w3[1+3*kk]; w3[2+3*indd1]=-w3[2+3*kk];        
    }

//   =============== STEP 5 = find maximum distance between points  ======================

    rr = 0;
    for(ii=0;ii<Nh;ii++) {
        for(kk=0;kk<(Nh-1);kk++) {
            ttn=0; for(jj=0;jj<3;jj++) ttn+= (w1[jj+3*(ii*Nh+kk)]-w1[jj+3*(ii*Nh+1+kk)])*(w1[jj+3*(ii*Nh+kk)]-w1[jj+3*(ii*Nh+1+kk)]); ttn=sqrt(ttn);
            if (ttn > rr)  rr = ttn; 
        }
    }
    for(kk=0;kk<Nh;kk++) {
        ttn= fabs(2.0*w1[1+3*kk]);
        if (ttn > rr)  rr = ttn; 
    }
    r_box[0]=rr;

//   =============== STEP 6 = find rectangular region containing all boxes from the quater-sector  ======================
// clean up
   free(x_sph); free(y_sph); free(z_sph);
   free(l_arc_reg); free(loc_vc); free(loc_vcc); free(dloc_vc); free(x_new); free(y_new);
   free(x_cube); free(y_cube);

   return 0;
}



long DBox::find_opposite(int scale, long snu)
{
long j, ind, Nangles = N_nu[scale];
double w11 = boxw1[scale][3*snu], w12 = boxw1[scale][1+3*snu], w13 = boxw1[scale][2+3*snu]; 
double prod, maxpr = - 1.0;

   for (j=0;j< Nangles;j++){
       prod= - w11*boxw1[scale][3*j] - w12*boxw1[scale][1+3*j] - w13*boxw1[scale][2+3*j];
       if(prod>maxpr) {maxpr=prod; ind=j; } else { }
   }
   printf("maxprod: %f\n",maxpr);
   return(ind);
}  


long DBox::find_closest(int scale, double *vect) // find orientation closest to a given vector
{
long j, ind, Nangles = N_nu[scale];
// double w11 = boxw1[scale][3*snu], w12 = boxw1[scale][1+3*snu], w13 = boxw1[scale][2+3*snu]; 
double prod, maxpr = - 1.0;

prod=sqrt(vect[0]*vect[0]+vect[1]*vect[1]+vect[2]*vect[2]); vect[0]=vect[0]/prod; vect[1]=vect[1]/prod; vect[2]=vect[2]/prod;

   for (j=0;j< Nangles;j++){
       prod=  vect[0]*boxw1[scale][3*j] +  vect[1]*boxw1[scale][1+3*j] +  vect[2]*boxw1[scale][2+3*j];
       if(prod>maxpr) {maxpr=prod; ind=j; } else { }
   }
//   printf("maxprod: %f\n",maxpr);
   return(ind);
} 






//*************************************************************************************//
// Window functions and grid points within box

const double Pi=3.14159265358979323846;
const double Pihalf=1.570796326794896558;

const int APOLP[10][11]={{3,-2},{10,-15,6},{35,-84,70,-20},{126,-420,540,-315,70},{462,-1980,3465,-3080,1386,-252},{1716,-9009,20020,-24024,16380,-6006,924},{6435,-40040,108108,-163800,150150,-83160,25740,-3432},{24310,-175032,556920,-1021020,1178100,-875160,408408,-109395,12870},{92378,-755820,2771340,-5969040,8314020,-7759752,4849845,-1956240,461890,-48620},{352716,-3233230,13430340,-33256080,54318264,-61108047,47927880,-25865840,9189180,-1939938,184756}};


void DBox::apol(double* t,double *y,int grSize){
  int j,k;
  double* s=(double*)malloc(grSize*sizeof(double));  double *tt=(double*)malloc(grSize*sizeof(double));
  for(j=0;j<grSize;j++){tt[j]=fmax(t[j],0);tt[j]=fmin(tt[j],1);s[j]=tt[j];y[j]=(double)APOLP[APOLDEG-1][0];}
  for (k=1;k<=APOLDEG;k++)
  for(j=0;j<grSize;j++) {y[j]+=APOLP[APOLDEG-1][k]*s[j]; s[j]*=tt[j]; }
  for(j=0;j<grSize;j++) y[j]*=s[j];
  for(j=0;j<grSize;j++) {y[j]=fmax(y[j],0.0); y[j]=fmin(y[j],1.0); }
  free(s); free(tt);
}

void DBox::meyer(double *t, double *y, int grSize,int type){
  /* INPUT:  t - array of values to compute meyer function on (grid as an array),
             grSize - size of t.
     OUTPUT: y - meyer function values for points in t.
  */
  int j;
  double *a=(double*)malloc(grSize*sizeof(double));
  double *b=(double*)malloc(grSize*sizeof(double));
  double tmp;

  switch(type){
  case MEYERCOARSE:
      for(j=0;j<grSize;j++) a[j]=2*fabs(t[j])-1;
      apol(a,b,grSize);
      for(j=0;j<grSize;j++) {tmp=Pihalf*b[j]; if(tmp >= Pihalf) y[j]=0; else y[j]=cos(tmp); }
      break;

  case MEYERDETAIL:
      for(j=0;j<grSize;j++) {
	 if(t[j]<=1.0) a[j]=2*t[j]-1; else a[j]=t[j]-1;   
      }
      apol(a,b,grSize);

      for(j=0;j<grSize;j++) {
         tmp=Pihalf*b[j]; 	
         if(t[j]<=1.0) { if(tmp <= 0) y[j]=0.0; else y[j]=sin(tmp); } 
         else { if(tmp > Pihalf) y[j]=0.0; else y[j]=cos(tmp); }
      }
      for(j=0;j<grSize;j++) { if((t[j]<=0.5)||(t[j]>=2.0)) y[j]=0.0; else ; }
      break;

  case MEYERFINE: // finest scale window has the same slope on the right as on the left
    for(j=0;j<grSize;j++) {
      if(t[j]<=1.0) a[j]=2.0*t[j]-1.0; else a[j]=2.0*t[j]-3.0;
    }
    apol(a,b,grSize);

    for(j=0;j<grSize;j++) {
      tmp=Pihalf*b[j];
      if(t[j]<=1.0) { if(tmp <= 0) y[j]=0.0; else y[j]=sin(tmp); }
      else { if(tmp > Pihalf) y[j]=0.0; else y[j]=cos(tmp); }
    }
    for(j=0;j<grSize;j++) { if((t[j]<=0.5)||(t[j]>=2.0)) y[j]=0.0; else ; }

  }
  free(a);free(b);    
}



void DBox::chihat_ang(float *x1, float *x2, float *x3, double *chihat, int *N, double *L, double Scale_rad){
   int j;
   double *rho=(double*)malloc(N[0]*N[1]*N[2]*sizeof(double));
   double *r=(double*)malloc(N[0]*N[1]*N[2]*sizeof(double));
   double *a=(double*)malloc(N[0]*N[1]*N[2]*sizeof(double));
   double tmp;

   for(j=0;j<N[0]*N[1]*N[2];j++) {
       rho[j]=fmin(sqrt(x2[j]*x2[j]+x3[j]*x3[j]),L[1]);
       r[j]=sqrt(x1[j]*x1[j]+x2[j]*x2[j]+x3[j]*x3[j]);
       a[j]=atan2(rho[j],x1[j])*r[j]/(atan(L[1]/(4*Scale_rad))*2*Scale_rad);   
   }
   apol(a,r,N[0]*N[1]*N[2]);
   for(j=0;j<N[0]*N[1]*N[2];j++) {tmp=Pihalf*r[j]; if(tmp >= Pihalf) chihat[j]=0; else chihat[j]=cos(tmp); }
  
   free(r);free(a);free(rho);
}



void DBox::chihat_rad(float *x1, float *x2, float *x3, double *chihat, int *N, double Scale_rad, int type){
  int j;
  // double nf=1.0/Scale_rad; // substitute division by miltiplication later
  double *r=(double*)malloc(N[0]*N[1]*N[2]*sizeof(double));
  for(j=0;j<N[0]*N[1]*N[2];j++) r[j]=sqrt(x1[j]*x1[j]+x2[j]*x2[j]+x3[j]*x3[j])/Scale_rad;
  meyer(r, chihat, N[0]*N[1]*N[2],type);
  free(r);
}

void DBox::chihat_sm_f(float *x1, float *x2, float *x3,  // the same as chihat but checks intersection of boxes
		  double *bboxw1, double *bboxw2, double *bboxw3, 
		  float *w, int *N, double *L, double Scale_rad, int Np,int scale, int snu)
{

  long j;
  double *ttw;
  ttw=(double*) malloc(N[0]*N[1]*N[2]*sizeof(double));

  chihat_sm(x1, x2, x3, bboxw1, bboxw2, bboxw3, ttw, N, L, Scale_rad, Np,scale, snu);
  for(j=0;j<N[0]*N[1]*N[2];j++) w[j]=(float)ttw[j];  

  free(ttw);
}

void DBox::chihat_sm(float *x1, float *x2, float *x3,  // the same as chihat but checks intersection of boxes
		  double *boxw1, double *boxw2, double *boxw3, 
		  double *w, int *N, double *L, double Scale_rad, int Np,int scale, int snu)
{
 // INPUT:
 // x1, x2, x3 grid points; N is the number of points in each direction of the grid.
 // boxw1, boxw2, boxw3 orientations of the boxes. Are to be stored as int [3][Np]. 
 // Np - number of box orientations
 // angid is the index of the window function to be computed.    
 // OUTPUT:
 // w is the window function values at the points x1,x2,x3 (oriented).

  int j, nnu, type;
  double sw1[3], sw2[3], sw3[3], v1[3], v2[3], v3[3];
  float *l1,*l2,*l3; 
  double *tmp, *tmp1, *ws;
  double alpha, beta;
  double m1=10, m2=0;
  double overs_norm;
 
  overs_norm=1/sqrt(OverSampleFac[0][scale]*OverSampleFac[1][scale]*OverSampleFac[2][scale]); 
 
  if( scale<(Kend-1) ) type=MEYERDETAIL; else type=MEYERFINE; 
  // printf("!!!!!!!!!!!K1[0]=%f, K1local=%f",K1[0], Scale_rad);

  for(j=0;j<3;j++) {
     sw1[j]=boxw1[j+snu*3]; sw2[j]=boxw2[j+snu*3]; sw3[j]=boxw3[j+snu*3];  
  }
  
  l1=(float*) malloc(N[0]*N[1]*N[2]*sizeof(float));
  l2=(float*) malloc(N[0]*N[1]*N[2]*sizeof(float));
  l3=(float*) malloc(N[0]*N[1]*N[2]*sizeof(float));
  tmp=(double*) malloc(N[0]*N[1]*N[2]*sizeof(double));
  tmp1=(double*) malloc(N[0]*N[1]*N[2]*sizeof(double));
  ws=(double*) malloc(N[0]*N[1]*N[2]*sizeof(double));
  for(j=0;j<N[0]*N[1]*N[2];j++) ws[j]=0;


  alpha=2*atan(L2[scale]/2/K1low[scale]); // maximum opening angle for intersecting boxes

  for(nnu=0; nnu < Np; nnu++) {
      beta=acos(boxw1[nnu*3]*sw1[0] + boxw1[1+nnu*3]*sw1[1] + boxw1[2+nnu*3]*sw1[2]);
      if(beta > alpha) {}
      else {
           for(j=0;j<3;j++) { v1[j]=0.0;  v2[j]=0.0;  v3[j]=0.0; } 
           for(j=0;j<3;j++) {
	  v1[0]+= boxw1[j+nnu*3]*sw1[j]; v1[1]+= boxw1[j+nnu*3]*sw2[j]; v1[2]+= boxw1[j+nnu*3]*sw3[j]; // calculate rotation matrix
                    v2[0]+= boxw2[j+nnu*3]*sw1[j]; v2[1]+= boxw2[j+nnu*3]*sw2[j]; v2[2]+= boxw2[j+nnu*3]*sw3[j];
                    v3[0]+= boxw3[j+nnu*3]*sw1[j]; v3[1]+= boxw3[j+nnu*3]*sw2[j]; v3[2]+= boxw3[j+nnu*3]*sw3[j];
           }
           for(j=0;j<N[0]*N[1]*N[2];j++) {
	  l1[j]= x1[j]*v1[0]+ x2[j]*v1[1]+ x3[j]*v1[2];
                    l2[j]= x1[j]*v2[0]+ x2[j]*v2[1]+ x3[j]*v2[2];
                    l3[j]= x1[j]*v3[0]+ x2[j]*v3[1]+ x3[j]*v3[2];
           }
           chihat_ang(l1,l2,l3,tmp,N,L, Scale_rad);
          for(j=0;j<N[0]*N[1]*N[2];j++) ws[j]+=tmp[j]*tmp[j];
       }
  }

   chihat_ang(x1,x2,x3,tmp1,N,L,Scale_rad); //compute the "angular part" of the window function
   chihat_rad(x1,x2,x3,tmp,N,Scale_rad,type); // compute the radial part of the window function. The last
   for(j=0;j<N[0]*N[1]*N[2];j++) 
      w[j]=(tmp1[j]*tmp[j]*overs_norm/sqrt(ws[j]+0.00000000000000000000000000000001)); // partition of unity in the angular direction. 

   free(l1);free(l2);free(l3);free(tmp);free(tmp1);free(ws);
   return;
}



void DBox::chihat_f(float *x1, float *x2, float *x3, 
		  double *boxw1, double *boxw2, double *boxw3, 
		  float *w, int *N, double *L, double Scale_rad, int Np, int kk, int snu)
{
 // INPUT:
 // x1, x2, x3 grid points; N is the number of points in each direction of the grid.
 // boxw1, boxw2, boxw3 orientations of the boxes. Are to be stored as int [3][Np]. 
 // Np - number of box orientations
 // kk - scale
 // snu - index of the window function to be computed.    
 // OUTPUT:
 // w is the window function values at the points x1,x2,x3 (oriented).

  int j, nnu, type;
  double sw1[3], sw2[3], sw3[3], v1[3], v2[3], v3[3];
  float *l1,*l2,*l3;
  double *tmp, *tmp1,*ws;
  double m1=10, m2=0;
  double overs_norm;

  l1=(float*) malloc(N[0]*N[1]*N[2]*sizeof(float));
  l2=(float*) malloc(N[0]*N[1]*N[2]*sizeof(float));
  l3=(float*) malloc(N[0]*N[1]*N[2]*sizeof(float));
  tmp1=(double*) malloc(N[0]*N[1]*N[2]*sizeof(double));
  tmp=(double*) malloc(N[0]*N[1]*N[2]*sizeof(double));
  ws=(double*) malloc(N[0]*N[1]*N[2]*sizeof(double));
  for(j=0;j<N[0]*N[1]*N[2];j++) ws[j]=0;

 
  overs_norm=1/sqrt(OverSampleFac[0][kk]*OverSampleFac[1][kk]*OverSampleFac[2][kk]); 
 
  if(kk!=0) {   //  scale k > 0
       if(kk < (Kend-1)) type=MEYERDETAIL; else type=MEYERFINE; 
       // printf("K1[0]=%f, K1local=%f",K1[0], Scale_rad);

      for(j=0;j<3;j++) {   
           sw1[j]=boxw1[j+snu*3]; sw2[j]=boxw2[j+snu*3]; sw3[j]=boxw3[j+snu*3];    
      }
  
      for(nnu=0; nnu < Np; nnu++) {
         for(j=0;j<3;j++) { v1[j]=0.0;  v2[j]=0.0;  v3[j]=0.0; } 
         for(j=0;j<3;j++) {
             v1[0]+= boxw1[j+nnu*3]*sw1[j];  v1[1]+= boxw1[j+nnu*3]*sw2[j];  v1[2]+= boxw1[j+nnu*3]*sw3[j]; // calculate rotation matrix
             v2[0]+= boxw2[j+nnu*3]*sw1[j];  v2[1]+= boxw2[j+nnu*3]*sw2[j];  v2[2]+= boxw2[j+nnu*3]*sw3[j];
             v3[0]+= boxw3[j+nnu*3]*sw1[j];  v3[1]+= boxw3[j+nnu*3]*sw2[j];  v3[2]+= boxw3[j+nnu*3]*sw3[j];
         }
         for(j=0;j<N[0]*N[1]*N[2];j++) {
             l1[j]= x1[j]*v1[0]+ x2[j]*v1[1]+ x3[j]*v1[2];
             l2[j]= x1[j]*v2[0]+ x2[j]*v2[1]+ x3[j]*v2[2];
             l3[j]= x1[j]*v3[0]+ x2[j]*v3[1]+ x3[j]*v3[2];
         }
         chihat_ang(l1,l2,l3,tmp,N,L, Scale_rad);
         for(j=0;j<N[0]*N[1]*N[2];j++) ws[j]+=tmp[j]*tmp[j];
     }  // for(nnu=0....

      chihat_ang(x1,x2,x3,tmp1,N,L,Scale_rad); //compute the "angular part" of the window function
      chihat_rad(x1,x2,x3,tmp,N,Scale_rad,type); // compute the radial part of the window function. The last
      for(j=0;j<N[0]*N[1]*N[2];j++) 
         w[j]=(float)(tmp1[j]*tmp[j]*overs_norm/sqrt(ws[j]+0.00000000000000000000000000000001)); // partition of unity in the angular direction. 
  } // if(kk!=0)... 
  else {     // scale 0
      chihat_rad(x1,x2,x3,tmp,N, K1[1],MEYERCOARSE); 
      for(j=0;j<N[0]*N[1]*N[2];j++)   w[j]=(float)(tmp[j]*overs_norm); 
 }

   free(l1); free(l2); free(l3); free(tmp); free(tmp1); free(ws);
   return;
}


int DBox::CalcWindowFunc_f(int k, int ix_angle, float *window_func)
{
  int j1, j2, j3;
  int N[3];  N[0]=Nk[0][k]; N[1]=Nk[1][k]; N[2]=Nk[2][k];
  long ngrids = N[0]*N[1]*N[2];
  float *x1,*x2,*x3;
  x1=(float*) malloc(ngrids*sizeof(float));
  x2=(float*) malloc(ngrids*sizeof(float));
  x3=(float*) malloc(ngrids*sizeof(float));

  double *p1, *p2, *p3;
  double step_x1, step_x2, step_x3, K2low;
  double L[2];

  CalcGrid(k, x1, x2, x3);

  if(k==0) { 
     p1 = &boxw1[1][0];   p2 = &boxw2[1][0];  p3 = &boxw3[1][0]; // do not use it (may be can skip)
     L[0] = L1[0]; L[1] = L2[0];     
     chihat_f(x1, x2, x3, p1, p2, p3, window_func, N, L, K1[1], N_nu[1], 0, 1); // actually used input: x1, x2, x3, window_func, N, K1[1], 0
     }
  else {
     p1 = &boxw1[k][0];   p2 = &boxw2[k][0];  p3 = &boxw3[k][0];
     L[0] = L1[k]; L[1] = L2[k];     

     if(L2[k] < (2*K1low[k]))
           { chihat_sm_f(x1, x2, x3, p1, p2, p3, window_func, N, L, K1[k], N_nu[k], k,ix_angle); }
     else
          {chihat_f(x1, x2, x3, p1, p2, p3, window_func, N, L, K1[k], N_nu[k], k, ix_angle); }
  } // else
 
  // clean up
   free(x1); free(x2); free(x3);

  return 0;
}






