#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sfftw.h>

#ifdef __PAR__
#include <mpi.h>
#include <sfftw_mpi.h>
#endif

#include <string.h>
#include "dmath.h"
#include "curvelet.h"




int DCurvelet::E2UBox(fftw_complex *fftBox, fftw_complex ***fftData, int *Pst, int *Pend, int k, int ix_angle)
{         // INPUT: fftData of image on global grid
          // OUTPUT: fftBox values on a box
  int m = US_OverSample;

  int hq = US_d_eps;
  int q = 1+2*US_d_eps; // overall width of gaussian bell
  double lam=US_lambda;

  int nxp=Nx, nyp=Ny,  nzp=Nz;
  int Nxm = m*nxp, Nym = m*nyp,  Nzm = m*nzp;
  int lx, ly, lz;

  double thi1, thi2, thi3, dx1, dx2, dx3;
  double tmp, tmp1, tmp2; 
  double Qjk;

  double *gxl = DMath::newD(3*(q+1));
  double *gym = &gxl[q+1];
  double *gzn = &gxl[2*(q+1)];

  int jx, jy, jz;
  int l1, l2, l3, l1_st, l2_st, l3_st, l1_fin, l2_fin, l3_fin;

  int N[3], Nst[3], Nend[3];
  N[0] = WindowData->Nk[0][k]; N[1] = WindowData->Nk[1][k]; N[2] = WindowData->Nk[2][k];
  Nst[0]=Pst[0]+Nxm/2; Nst[1]=Pst[1]+Nym/2; Nst[2]=Pst[2]+Nzm/2; Nend[0]=Pend[0]+Nxm/2; Nend[1]=Pend[1]+Nym/2; Nend[2]=Pend[2]+Nzm/2;

  long il, ngrid=N[0]*N[1]*N[2];

   float *xx1, *xx2, *xx3;
   xx1=(float*) malloc(ngrid*sizeof(float));  xx2=(float*) malloc(ngrid*sizeof(float));  xx3=(float*) malloc(ngrid*sizeof(float));

  WindowData->CalcGrid(k, ix_angle, xx1, xx2, xx3);

  long Nt = nxp*nyp*nzp;
  double temp_v, norm_fac;

  norm_fac=sqrt(lam/(double)PI);  // normalization factor for USFFT (sqrt(lambda/PI))^3 for FFT it is 1/sqrt(Nt)
  norm_fac=norm_fac*norm_fac*norm_fac/sqrt((double)(Nt)); 

  for(il=0; il < ngrid; il++) {
 
       thi1 = xx1[il]*(double)m + Nxm/2;  thi2 = xx2[il]*(double)m + Nym/2; thi3 = xx3[il]*(double)m + Nzm/2; 

       l1 = thi1>0 ? (int)(thi1+0.5) : (int)(thi1-0.5); // nearest integer
       l2 = thi2>0 ? (int)(thi2+0.5) : (int)(thi2-0.5);
       l3 = thi3>0 ? (int)(thi3+0.5) : (int)(thi3-0.5);
  
       l1_st=max((l1-hq), Nst[0])-(l1-hq); l1_fin=min((l1+hq), (Nend[0]-1))+1-(l1-hq);  if( ((l1-hq) < Nst[0])&&(Nst[0]>(1)) )  cout << "TROUUUUUBLEEEEEE !!!!!!!! 1.1: "<< Nst[0]-(l1-hq) << "\n"; cout.flush();  if( ((l1+hq)>(Nend[0]-1))&&(Nend[0]<(-1)) )  cout << "TROUUUUUBLEEEEEE !!!!!!!! 1.2: "<< (l1+hq)-(Nend[0]-1) << ", Nend[0]: "<< Nend[0] <<"\n"; cout.flush(); 
       l2_st=max((l2-hq), Nst[1])-(l2-hq); l2_fin=min((l2+hq), (Nend[1]-1))+1-(l2-hq);  if( ((l2-hq) < Nst[1])&&(Nst[1]>(1)) )  cout << "TROUUUUUBLEEEEEE !!!!!!!! 2.1: Pst="<< Nst[1] << ", whatever= " << (l2-hq) << "\n"; cout.flush();  if( ((l2+hq)>(Nend[1]-1))&&(Nend[1]<(-1)) )  cout << "TROUUUUUBLEEEEEE !!!!!!!! 2.2: "<< (l2+hq)-(Nend[1]-1) << ", Nend[1]: "<< Nend[1] << "\n"; cout.flush(); 
       l3_st=max((l3-hq), Nst[2])-(l3-hq); l3_fin=min((l3+hq), (Nend[2]-1))+1-(l3-hq);   
          
       dx1=2.0*lam*(thi1-(double)l1);   dx2=2.0*lam*(thi2-(double)l2);   dx3=2.0*lam*(thi3-(double)l3);

      fftBox[il].re = fftBox[il].im = 0.0;

       get_gauss_bell(thi1-l1, l1_st, l1_fin, gxl);
       get_gauss_bell(thi2-l2, l2_st, l2_fin, gym);
       get_gauss_bell(thi3-l3, l3_st, l3_fin, gzn); 

      for(jx=l1_st; jx<l1_fin; jx++) {
              lx =(l1-hq + jx) - Nst[0];  // if( (lx < 0) || (lx > (Nxm-1)) )  cout << "TROUUUUUBLEEEEEE !!!!!!!!  3\n"; cout.flush(); 

              for(jy=l2_st; jy<l2_fin;jy++) {
	   ly = (l2-hq + jy) - Nst[1];  // if( (ly < 0) || (ly > (Nym-1)) )  cout << "TROUUUUUBLEEEEEE !!!!!!!!  3\n"; cout.flush(); 
	   temp_v = gxl[jx]*gym[jy];

	   for(jz=l3_st; jz<l3_fin; jz++) {
                          lz = (l3-hq + jz) - Nst[2];  //  if( (lz < 0) || (lz > (Nzm-1)) )   cout << "TROUUUUUBLEEEEEE !!!!!!!!  3\n"; cout.flush(); 
                          Qjk = temp_v*gzn[jz];
                             
 	        fftBox[il].re += Qjk*fftData[lx][ly][lz].re;  fftBox[il].im += Qjk*fftData[lx][ly][lz].im;
	   } // for(jz=0...
             } // for(jy=0...
     } // for(jx=0...

    fftBox[il].re*=norm_fac;  fftBox[il].im*=norm_fac;

  } // for il=0

  DMath::del(gxl);
  free(xx1);  free(xx2);  free(xx3); 

  return 0;
}


int DCurvelet::U2EBox(fftw_complex ***fftData, fftw_complex *fftBox, int *Pst, int *Pend, int k, int ix_angle, int flag_addimage)
{         // OUTPUT: fftData of image on global grid
          // INPUT: fftBox values on a box

  int m = US_OverSample;
  int nxp=Nx, nyp=Ny, nzp=Nz;
  int Nxm = m*nxp, Nym = m*nyp, Nzm = m*nzp;

  int hq = US_d_eps;
  int q = 1+2*US_d_eps; // overall width of gaussian bell
  double lam=US_lambda;

  if(flag_addimage==0)  memset(&fftData[0][0][0], '\0', sizeof(fftw_complex)*(Nxm)*(Nym)*(Nzm));

  double thi1, thi2, thi3, dx1, dx2, dx3;
  double Pjk , temp_v;

  double *gxl = DMath::newD(3*(q+1));
  double *gym = &gxl[q+1];
  double *gzn = &gxl[2*(q+1)];

  int N[3], Nst[3], Nend[3];
  N[0] = WindowData->Nk[0][k]; N[1] = WindowData->Nk[1][k]; N[2] = WindowData->Nk[2][k];
  Nst[0]=Pst[0]+Nxm/2; Nst[1]=Pst[1]+Nym/2; Nst[2]=Pst[2]+Nzm/2; Nend[0]=Pend[0]+Nxm/2; Nend[1]=Pend[1]+Nym/2; Nend[2]=Pend[2]+Nzm/2;

  long il, ngrid=N[0]*N[1]*N[2];
  int lx, ly, lz, jx, jy, jz;
  int l1, l2, l3, l1_st, l2_st, l3_st, l1_fin, l2_fin, l3_fin;

 
    float *xx1, *xx2, *xx3;
    xx1=(float*) malloc(ngrid*sizeof(float));
    xx2=(float*) malloc(ngrid*sizeof(float));
    xx3=(float*) malloc(ngrid*sizeof(float));

  WindowData->CalcGrid(k, ix_angle, xx1, xx2, xx3);

  long Nt = nxp*nyp*nzp;
  double norm_fac;

  for(il=0; il < ngrid; il++) {
       thi1 = xx1[il]*(double)m+ Nxm/2;  thi2 = xx2[il]*(double)m+ Nym/2; thi3 = xx3[il]*(double)m+ Nzm/2; 

       l1 = thi1>0 ? (int)(thi1+0.5) : (int)(thi1-0.5); // nearest integer
       l2 = thi2>0 ? (int)(thi2+0.5) : (int)(thi2-0.5);
       l3 = thi3>0 ? (int)(thi3+0.5) : (int)(thi3-0.5);

       l1_st=max((l1-hq), Nst[0])-(l1-hq); l1_fin=min((l1+hq), (Nend[0]-1))+1-(l1-hq);  if( ((l1-hq) < Nst[0])&&(Nst[0]>(1)) )  cout << "TROUUUUUBLEEEEEE !!!!!!!! 1.1: "<< Nst[0]-(l1-hq) << "\n"; cout.flush();  if( ((l1+hq)>(Nend[0]-1))&&(Nend[0]<(-1)) )  cout << "TROUUUUUBLEEEEEE !!!!!!!! 1.2: "<< (l1+hq)-(Nend[0]-1) << ", Nend[0]: "<< Nend[0] <<"\n"; cout.flush(); 
       l2_st=max((l2-hq), Nst[1])-(l2-hq); l2_fin=min((l2+hq), (Nend[1]-1))+1-(l2-hq);  if( ((l2-hq) < Nst[1])&&(Nst[1]>(1)) )  cout << "TROUUUUUBLEEEEEE !!!!!!!! 2.1: Pst="<< Nst[1] << ", whatever= " << (l2-hq) << "\n"; cout.flush();  if( ((l2+hq)>(Nend[1]-1))&&(Nend[1]<(-1)) )  cout << "TROUUUUUBLEEEEEE !!!!!!!! 2.2: "<< (l2+hq)-(Nend[1]-1) << ", Nend[1]: "<< Nend[1] << "\n"; cout.flush(); 
       l3_st=max((l3-hq), Nst[2])-(l3-hq); l3_fin=min((l3+hq), (Nend[2]-1))+1-(l3-hq);   
 
       get_gauss_bell(thi1-l1, l1_st, l1_fin, gxl); 
       get_gauss_bell(thi2-l2, l2_st, l2_fin, gym); 
       get_gauss_bell(thi3-l3, l3_st, l3_fin, gzn); 

       for(jx=l1_st; jx<l1_fin; jx++) {
                 lx =(l1-hq + jx) - Nst[0];
   //              lx = ((l1-hq + jx) + Nxm)% Nxm;
        
                 for(jy=l2_st; jy<l2_fin;jy++) {
	   ly = (l2-hq + jy) - Nst[1]; 
 //	   ly = ((l2-hq + jy) + Nym) % Nym;
	   temp_v = gxl[jx]*gym[jy];

	   for(jz=l3_st; jz<l3_fin; jz++) {
                         lz = (l3-hq + jz) - Nst[2];  
//                       lz = ((l3-hq + jz) + Nzm) % Nzm;
                         Pjk = temp_v*gzn[jz];

  	      fftData[lx][ly][lz].re += fftBox[il].re*Pjk;  fftData[lx][ly][lz].im += fftBox[il].im*Pjk;
	   } // for(jz=0...
                } // for(jy=0...
       } // for(jx=0 ...

    } // for il=0

  DMath::del(gxl);
  free(xx1);  free(xx2);  free(xx3); 

  return 0;
}


int DCurvelet::seqFFT(fftw_complex ***fftData, fftw_complex ***func) 
{

  int m = US_OverSample;
  int nxp=Nx, nyp=Ny,  nzp=Nz;
  int Nxm = m*nxp, Nym = m*nyp,  Nzm = m*nzp;

  int i, j, kk, lx, ly, lz;
  long lind;
  double temp_v, temp_v1;

  double *gjx = DMath::newD(nxp+nyp+nzp);
  double *gjy = &gjx[nxp];
  double *gjz = &gjx[nxp+nyp];

  get_normal_coeff(m, nxp, gjx);   get_normal_coeff(m, nyp, gjy);   get_normal_coeff(m, nzp, gjz);

  fftw_complex ***func_scaled = DMath::newC(Nxm, Nym, Nzm);
  memset(&func_scaled[0][0][0], '\0', sizeof(fftw_complex)*Nxm*Nym*Nzm);

  for(i=0;i<nxp;i++) // first half in x
    {
      if(i<nxp/2) lx = Nxm - nxp/2 + i;  else lx = i - nxp/2;

      for(j=0;j<nyp;j++) // first half in y
	{
	  if(j<nyp/2) ly = Nym - nyp/2 + j;  else ly = j - nyp/2;
	  temp_v = gjx[i]*gjy[j];

	  for(kk=0;kk<nzp/2;kk++) {
	     lz = Nzm - nzp/2 + kk;
	     lind=(lx+ly+lz) % 2; if (lind==0) {lind=1;} else {lind=-1;}  // alternating 1 and -1 to get zero freq. in the middle
                       temp_v1 = temp_v*gjz[kk]*lind;         //temp_v1 = temp_v*gjz[kk];
  	     func_scaled[lx][ly][lz].re = func[i][j][kk].re*temp_v1;  func_scaled[lx][ly][lz].im = func[i][j][kk].im*temp_v1;
	  } // for(k=0...

	  for(kk=nzp/2;kk<nzp;kk++) {
 	     lz = kk - nzp/2;
	     lind=(lx+ly+lz) % 2; if (lind==0) {lind=1;} else {lind=-1;}  // alternating 1 and -1 to get zero freq. in the middle
	     temp_v1 = temp_v*gjz[kk]*lind;   //temp_v1 = temp_v*gjz[kk];
	     func_scaled[lx][ly][lz].re = func[i][j][kk].re*temp_v1;  func_scaled[lx][ly][lz].im = func[i][j][kk].im*temp_v1;
	  } // for(kk=nzp/2...

	} // for(j=0...
    } // for(i=0...


  fftwnd_plan plan_forward = fftw3d_create_plan(Nxm, Nym, Nzm, FFTW_FORWARD, FFTW_ESTIMATE);
  fftwnd(plan_forward, 1, &func_scaled[0][0][0], 1, Nxm*Nym*Nzm, &fftData[0][0][0], 1, Nxm*Nym*Nzm);
  fftwnd_destroy_plan(plan_forward);

  DMath::del(func_scaled);
  DMath::del(gjx);

  return 0;
}


 int DCurvelet::seqIFFT(fftw_complex ***func, fftw_complex ***fftDat) 
{
  int m = US_OverSample;
  int nxp=Nx, nyp=Ny, nzp=Nz;
  int Nxm = m*nxp, Nym = m*nyp, Nzm = m*nzp;
   
  int i, j, kk, lx, ly, lz;
  long lind;
  
 for(i=0;i<Nxm;i++) 
      for(j=0;j<Nym;j++) 
          for(kk=0;kk<Nzm;kk++) {
               lind=(i+j+kk) % 2; if (lind==0) {lind=1;} else {lind=-1;}  // alternating 1 and -1 to get centered image
               fftDat[i][j][kk].re*= lind;  fftDat[i][j][kk].im*= lind;
           } // for(kk=...


  double temp_v, temp_v1;
  
  fftwnd_plan plan_inv = fftw3d_create_plan(Nxm, Nym, Nzm, FFTW_BACKWARD, FFTW_ESTIMATE);
  fftw_complex ***imgData = DMath::newC(Nxm, Nym, Nzm);
  fftwnd(plan_inv, 1, &fftDat[0][0][0], 1, 0, &imgData[0][0][0], 1, 0);
  fftwnd_destroy_plan(plan_inv);


  double *gjx = DMath::newD(nxp+nyp+nzp);
  double *gjy = &gjx[nxp];
  double *gjz = &gjx[nxp+nyp];

  get_normal_coeff(m, nxp, gjx);
  get_normal_coeff(m, nyp, gjy);
  get_normal_coeff(m, nzp, gjz);

  long Nt = nxp*nyp*nzp;
  double lam=US_lambda;
  double norm_fac;
  
  norm_fac=sqrt(lam/(double)PI);  // normalization factor for USFFT (sqrt(lambda/PI))^3 for FFT it is 1/sqrt(Nt)
  norm_fac=norm_fac*norm_fac*norm_fac/sqrt((double)(Nt)); 

for(i=0;i<nxp;i++)   {
    lx = i + Nxm/2 - nxp/2;
    for(j=0;j<nyp;j++)    {
       ly = j + Nym/2  - nyp/2;
       temp_v = gjx[i]*gjy[j];
           for(kk=0;kk<nzp;kk++) {
                lz = kk + Nzm/2  - nzp/2;
                temp_v1 = temp_v*gjz[kk];
                func[i][j][kk].re = imgData[lx][ly][lz].re*temp_v1*norm_fac;
                func[i][j][kk].im = imgData[lx][ly][lz].im*temp_v1*norm_fac;
           } // for(kk=...
      } // for(j=0...
  } // for(i=0...


  DMath::del(imgData);
  DMath::del(gjx);

  return 0;
}


void DCurvelet::get_normal_coeff(int m, int N, double *gjx)
{
  int ix;
  double logeps2 = -2*log(US_Epsilon)/(double)(m*N*m*N);
  for(ix=0; ix<N; ix++) {
    gjx[ix] = exp(logeps2*(ix-N/2)*(ix-N/2));
  }
}
void DCurvelet::get_normal_coeff(int m, int N, int Nst, int Nend, double *gjx)
{
  int ix;
  double logeps2 = -2*log(US_Epsilon)/(double)(m*N*m*N);
  for(ix=Nst; ix<Nend; ix++) {
    gjx[ix-Nst] = exp(logeps2*(ix-N/2)*(ix-N/2));
  }
}

void DCurvelet::get_gauss_bell(double xi, int q, double *gxl)
{
  int i;
  double dxi;
  for(i=0;i<q;i++) { dxi=xi-(i-US_d_eps);  gxl[i]=exp(-US_lambda*dxi*dxi); }
}

void DCurvelet::get_gauss_bell(double xi, int q1, int q2, double *gxl)
{
  int i;
  double dxi;
  for(i=q1;i<q2;i++) { dxi=xi-(i-US_d_eps);  gxl[i]=exp(-US_lambda*dxi*dxi);  }
}





#ifdef __PAR__

int DCurvelet::parFFT(char *input_file, char *max_v_file, MPI_Comm comm) 
{
  int fftmy_rank, fftncpus, sendcpu, reccpu;
  int local_nx, local_x_start, local_ny_after_transpose, local_y_start_after_transpose, total_local_size;

  MPI_Comm_rank(comm, &fftmy_rank);
  MPI_Comm_size(comm, &fftncpus);

  MPI_Status status;
  int m = US_OverSample;
  int Nxm = m*Nx, Nym = m*Ny,  Nzm = m*Nz;

  int lNx, indNx, num_read; 
            lNx=Nxm/fftncpus; indNx=lNx*fftmy_rank; // lNx, indNx- local length and starting index in x direction, 
            num_read=Nx/lNx; // number of chunks to read from input file (fftncpus/2)
            sendcpu=fftncpus/2;
  int ind, iter, i, j, kk, lx, ly, lz;
  long lind;
  double max_v, max_vv, tmp, tmp1, pi=3.14159265;
  FILE *fid;

  float *read_buf;  read_buf=(float*)malloc(sizeof(float)*lNx*Ny*Nz);
 
 fftw_complex ***func_scaled = DMath::newC(lNx, Nym, Nzm);
  memset(&func_scaled[0][0][0], '\0', sizeof(fftw_complex)*lNx*Nym*Nzm);

  double *gjx, *gjy, *gjz;
       gjx = DMath::newD(lNx); gjy = DMath::newD(Ny);  gjz = DMath::newD(Nz);


   if(fftmy_rank==sendcpu) {
      max_v=0.0;
      for(iter=0;iter<num_read;iter++) {
          if(iter<num_read/2)  reccpu = fftncpus-num_read/2+iter;  else reccpu = iter-num_read/2;
 
          fid = fopen(input_file,"rb");
          fseek (fid , iter*(lNx*Ny*Nz) *sizeof(float), SEEK_SET);
          fread(&read_buf[0], sizeof(float), lNx*Ny*Nz, fid);
          fclose (fid);
    cout << "iter:"<< iter << ", send from:"<< fftmy_rank << ", to reccpu=" << reccpu <<"\n"; cout.flush();        
          for(i=0;i<(lNx*Ny*Nz);i++)  if(fabs(read_buf[i])>max_v)  max_v=fabs(read_buf[i]);
if(1<0){
          for(i=0;i<lNx;i++)   for(j=0;j<Ny;j++)  for(kk=0;kk<8;kk++) {  
                       read_buf[kk+Ny*(j+lNx*i)]=(float)(read_buf[kk+Ny*(j+lNx*i)]*sin(kk*pi/16.0));
                       read_buf[Nz-kk+Ny*(j+lNx*i)]=(float)(read_buf[Nz-kk+Ny*(j+lNx*i)]*sin(kk*pi/16.0));
           }
          for(i=0;i<lNx;i++)   for(j=0;j<8;j++)  for(kk=0;kk<Nz;kk++) {  
                       read_buf[kk+Ny*(j+lNx*i)]=(float)(read_buf[kk+Ny*(j+lNx*i)]*sin(j*pi/16.0));
                       read_buf[kk+Ny*(Ny-j+lNx*i)]=(float)(read_buf[kk+Ny*(Ny-j+lNx*i)]*sin(j*pi/16.0));
           }
           if(iter==0)
               for(i=0;i<8;i++)   for(j=0;j<Ny;j++)  for(kk=0;kk<Nz;kk++)   read_buf[kk+Ny*(j+lNx*i)]=(float)(read_buf[kk+Ny*(j+lNx*i)]*sin(i*pi/16.0));
           if(iter==(num_read-1))
               for(i=0;i<8;i++)   for(j=0;j<Ny;j++)  for(kk=0;kk<Nz;kk++)   read_buf[kk+Ny*(j+lNx*(lNx-i))]=(float)(read_buf[kk+Ny*(j+lNx*(lNx-i))]*sin(i*pi/16.0));
}         
           MPI_Send(&read_buf[0], lNx*Ny*Nz, MPI_FLOAT, reccpu, 99, comm);
      } // for(iter=0;iter<nu...

      fid = fopen(max_v_file,"wb");  fwrite(&max_v, sizeof(double), 1, fid);  fclose (fid);      
      cout << "max_v=" << max_v <<"\n"; cout.flush();
   } // if (fftmy_rank=...

  if((fftmy_rank<num_read/2)||(fftmy_rank>(fftncpus - num_read/2-1))) {
        if(fftmy_rank<num_read/2)  ind=fftmy_rank+num_read/2;  else  ind=fftmy_rank-(fftncpus-num_read/2);
 cout << "cpu:" << fftmy_rank << " got data as: "<< ind <<  "\n"; cout.flush();
        MPI_Recv(&read_buf[0], lNx*Ny*Nz, MPI_FLOAT, sendcpu, 99, comm, &status);            
   }

   get_normal_coeff(m, Ny, 0, Ny, gjy);  get_normal_coeff(m, Nz, 0, Nz, gjz); 
   MPI_Bcast (&max_v, 1, MPI_DOUBLE, sendcpu,  comm); 
   max_vv=1.0/max_v;


 if((fftmy_rank<num_read/2)||(fftmy_rank>(fftncpus - num_read/2-1))) {
        if(fftmy_rank<num_read/2)  ind=fftmy_rank+num_read/2;  else  ind=fftmy_rank-(fftncpus-num_read/2);

        get_normal_coeff(m, Nx, ind*lNx, (ind+1)*lNx, gjx);  

        for(i=0;i<lNx;i++) // first half in x
            {
            lx=i;
            for(j=0;j<Ny;j++) // first half in y
//                 tmp = gjx[i]*max_v;
                {
                 if(j<Ny/2)  ly = Nym - Ny/2 + j;  else ly = j - Ny/2;
                 tmp = gjx[i]*gjy[j]*max_vv;
                 for(kk=0;kk<Nz/2;kk++) {
                      lz = Nzm - Nz/2 + kk;  ind=kk+Nz*(j+Ny*i); 
	    lind=(lx+ly+lz) % 2; if (lind==0) {lind=1;} else {lind=-1;}  // alternating 1 and -1 to get zero freq. in the middle
                      tmp1 = tmp*gjz[kk]*lind;         //temp_v1 = temp_v*gjz[kk];
  	    func_scaled[lx][ly][lz].re = read_buf[ind]*tmp1;  
//  	    func_scaled[lx][ly][lz].re = read_buf[ind]*lind;  
	} 
                  for(kk=Nz/2;kk<Nz;kk++) {
   	    lz = kk - Nz/2;  ind=kk+Nz*(j+Ny*i); 
	    lind=(lx+ly+lz) % 2; if (lind==0) {lind=1;} else {lind=-1;}  // alternating 1 and -1 to get zero freq. in the middle
	    tmp1 = tmp*gjz[kk]*lind;   // temp_v1 = tmp*gjz[kk];
  	    func_scaled[lx][ly][lz].re =read_buf[ind]*tmp1;  
//	    func_scaled[lx][ly][lz].re = read_buf[ind]*lind;  
	} 
            } // for(j=0...
        } // for(i=0...
  } // if((fftmy_rank<...

  free(read_buf);

  fftwnd_mpi_plan  plan_forward;
                      plan_forward= fftw3d_mpi_create_plan(comm, Nxm, Nym, Nzm, FFTW_FORWARD, FFTW_ESTIMATE);
  fftwnd_mpi(plan_forward, 1, &func_scaled[0][0][0], NULL, FFTW_NORMAL_ORDER);
  fftwnd_mpi_destroy_plan(plan_forward);

  long Nt = Nx*Ny*Nz;
  double lam=US_lambda;
  double norm_fac;
  
//  norm_fac=sqrt(lam/(double)PI);  // normalization factor for USFFT (sqrt(lambda/PI))^3 for FFT it is 1/sqrt(Nt)
//  norm_fac=norm_fac*norm_fac*norm_fac/sqrt((double)(Nt)); 

     // write Fourier spectrum into file 
//  for(i=0;i<lNx;i++)   for(j=0;j<Nym;j++)    for(kk=0;kk<Nzm;kk++) {
//            func_scaled[i][j][kk].re*=norm_fac; func_scaled[i][j][kk].im*=norm_fac; }

  if(fftmy_rank!=0)   MPI_Send(&func_scaled[0][0][0], 2*lNx*Nym*Nzm, MPI_FLOAT, 0, 99, comm);
  if(fftmy_rank==0) {
      fid = fopen("fft_spectr.in","wb");  fwrite(&func_scaled[0][0][0], sizeof(fftw_complex), lNx*Nym*Nzm, fid);  fclose (fid);
      for(iter=1; iter<fftncpus; iter++) {
           MPI_Recv(&func_scaled[0][0][0], 2*lNx*Nym*Nzm, MPI_FLOAT, iter, 99, comm, &status);            
           fid = fopen("fft_spectr.in","ab");  fwrite(&func_scaled[0][0][0], sizeof(fftw_complex), lNx*Nym*Nzm, fid);  fclose (fid);
      } // for(iter=0;iter<nu...
  } // if(fftmy_... 


  DMath::del(func_scaled);
  DMath::del(gjx);  DMath::del(gjy);  DMath::del(gjz);
//  free(write_buf);

  return 0;
}


int DCurvelet::parIFFT(char *output_file, char *max_v_file, MPI_Comm comm) 
{
  int fftmy_rank, fftncpus, wrcpu, reccpu;

  MPI_Comm_rank(comm, &fftmy_rank);
  MPI_Comm_size(comm, &fftncpus);

  MPI_Status status;
  int m = US_OverSample;
  int Nxm = m*Nx, Nym = m*Ny,  Nzm = m*Nz;

  int lNx, indNx, num_read; 
            lNx=Nxm/fftncpus; indNx=lNx*fftmy_rank; // lNx, indNx- local length and starting index in x direction, 
            num_read=Nx/lNx; // number of chunks to read from input file (fftncpus/2)

  int ind, iter, i, j, kk, lx, ly, lz;
  long lind;
  double vvv, vvv1, max_v, tmp, tmp1;
  FILE *fid;

  float *write_buf, *read_buf; 

 fftw_complex ***func_scaled = DMath::newC(lNx, Nym, Nzm);

  if(fftmy_rank==0) {
      fid = fopen(max_v_file,"rb");  fread(&max_v, sizeof(double), 1, fid);  fclose (fid);
  }
  MPI_Bcast (&max_v, 1, MPI_DOUBLE, 0,  comm); 


  if(fftmy_rank==0) {
      for(iter=1;iter<fftncpus;iter++) {
          ind = (iter+fftncpus/2)%fftncpus;
          fid = fopen(OUTP_FFT_FILE,"rb");  
          fseek (fid , ind*(lNx*Nym*Nzm) *sizeof(fftw_complex), SEEK_SET);
          fread(&func_scaled[0][0][0], sizeof(fftw_complex), lNx*Nym*Nzm, fid);
          fclose (fid);
          MPI_Send(&func_scaled[0][0][0], 2*lNx*Nym*Nzm, MPI_FLOAT, iter, 99, comm);
      } // for(iter=0;iter<nu...
      ind = fftncpus/2;
      fid = fopen(OUTP_FFT_FILE,"rb");   
      fseek (fid , ind*(lNx*Nym*Nzm)*sizeof(fftw_complex), SEEK_SET);
      fread(&func_scaled[0][0][0], sizeof(fftw_complex), lNx*Nym*Nzm, fid);  
      fclose (fid);
  } // if (fftmy_rank=...
  if(fftmy_rank!=0)   MPI_Recv(&func_scaled[0][0][0], 2*lNx*Nym*Nzm, MPI_FLOAT, 0, 99, comm, &status);            


           // set the spectrum  with fftshift
      int nnn[2]; nnn[0]=Nym; nnn[1]=Nzm;
      for(i=0;i<lNx;i++)   fftshift2(&func_scaled[i][0][0], nnn, 0); 

      for(lx=0;lx<lNx;lx++)    for(ly=0;ly<Nym;ly++)      for(lz=0;lz<Nzm;lz++) {
                        lind=(lx+ly+lz) % 2; if (lind==0) {lind=1;} else {lind=-1;}  // alternating 1 and -1 to get zero freq. in the middle
                        func_scaled[lx][ly][lz].re*=lind; func_scaled[lx][ly][lz].im*=lind;   } 
   
                 //  IFFT itself
  fftwnd_mpi_plan  plan_inverse;
                      plan_inverse= fftw3d_mpi_create_plan(comm, Nxm, Nym, Nzm, FFTW_BACKWARD, FFTW_ESTIMATE);
  fftwnd_mpi(plan_inverse, 1, &func_scaled[0][0][0], NULL, FFTW_NORMAL_ORDER);
  fftwnd_mpi_destroy_plan(plan_inverse);

  double *gjx, *gjy, *gjz;
       gjx = DMath::newD(lNx); gjy = DMath::newD(Ny);  gjz = DMath::newD(Nz);
 
  long Nt = Nx*Ny*Nz;
  double lam=US_lambda;
  double norm_fac;
  
  norm_fac=sqrt(lam/(double)PI);  // normalization factor for USFFT (sqrt(lambda/PI))^3 for FFT it is 1/sqrt(Nt)
  norm_fac=norm_fac*norm_fac*norm_fac/sqrt((double)(Nt))*max_v; 
//  norm_fac=1/sqrt((double)(Nt)); 

  read_buf=(float*)malloc(sizeof(float)*lNx*Ny*Nz);
  memset(&read_buf[0], '\0', sizeof(float)*lNx*Ny*Nz);

   wrcpu=fftncpus/2-num_read/2;

   if( (fftmy_rank>(wrcpu-1))&&(fftmy_rank<(wrcpu+num_read)) ) {

        ind=fftmy_rank-wrcpu; 
        get_normal_coeff(m, Ny, 0, Ny, gjy);  get_normal_coeff(m, Nz, 0, Nz, gjz); 
        get_normal_coeff(m, Nx, ind*lNx, (ind+1)*lNx, gjx);  
         
        for(i=0;i<lNx;i++) {  
            lx=i;
            for(j=0;j<Ny;j++)  {
                 ly = Nym/2 - Ny/2 + j;  
                 tmp = gjx[i]*gjy[j];
                 for(kk=0;kk<Nz;kk++) {
                      lz = Nzm/2 - Nz/2 + kk;  lind=kk+Nz*(j+Ny*i); 
                      tmp1 = tmp*gjz[kk];
                      read_buf[lind]=2.0*func_scaled[lx][ly][lz].re*norm_fac*tmp1;  
 //                    read_buf[lind]=func_scaled[lx][ly][lz].re*norm_fac;  
                 } // for(kk=0...
            } // for(j=0...
        } // for(i=0...

      if(fftmy_rank!=wrcpu)   MPI_Send(&read_buf[0], lNx*Ny*Nz, MPI_FLOAT, wrcpu, 99, comm);
      if(fftmy_rank==wrcpu) {

          fid = fopen(output_file,"wb");  fwrite(&read_buf[0], sizeof(float), lNx*Ny*Nz, fid);  fclose (fid);
          for(iter=(wrcpu+1); iter<(wrcpu+num_read); iter++) {
cout << "cpu:" << fftmy_rank << ", getting data from:" << iter << "\n"; cout.flush();
                MPI_Recv(&read_buf[0], lNx*Ny*Nz, MPI_FLOAT, iter, 99, comm, &status);            
                fid = fopen(output_file,"ab");  fwrite(&read_buf[0], sizeof(float), lNx*Ny*Nz, fid);  fclose (fid);
          } // for(iter=0;iter<nu...
      } // if(fftmy_... 

  } // if( (fftmy_rank>...

  DMath::del(func_scaled);
  DMath::del(gjx);  DMath::del(gjy);  DMath::del(gjz);
  free(read_buf);

  return 0;
}

int DCurvelet::parBcastRedMem(fftw_complex ***block, int dNx, int dNy, int dNz, MPI_Comm comm)
  {
   int my_rank, ncpus;  MPI_Comm_rank(comm, &my_rank);  MPI_Comm_size(comm, &ncpus);
   int res, num, it;
   int bufl=(int)((float)(MAX_REDUCE_BUF_SIZE*1024*1024)/((float)(dNy*dNz*4*2)));

   if(bufl>dNx) {bufl=dNx; res=0; num=1;} else {res=dNx%bufl; num=dNx/bufl;}
   for(it=0;it<num;it++)  MPI_Bcast(&block[it*bufl][0][0], 2*(bufl*dNy*dNz), MPI_FLOAT, (ncpus-1), comm); 
   if(res>0)  MPI_Bcast(&block[num*bufl][0][0], 2*(res*dNy*dNz), MPI_FLOAT, (ncpus-1), comm); 


   return 0;
}


int DCurvelet::parReduceRedMem(fftw_complex ***imag_rec, int dNx, int dNy, int dNz, MPI_Comm comm)
  {
   int my_rank, ncpus;  MPI_Comm_rank(comm, &my_rank);  MPI_Comm_size(comm, &ncpus);
   int res, num, ii, jj, kk, ll, it;
   int bufl=(int)((float)(MAX_REDUCE_BUF_SIZE*1024*1024/4/2)/((float)(dNy*dNz)));
    
   if(bufl>dNx) {bufl=dNx; res=0; num=1;} else { res=dNx%bufl; num=dNx/bufl;}
   fftw_complex *buf_write = DMath::newC(bufl*dNy*dNz);
   for(it=0; it<num; it++)  {
        MPI_Reduce(&imag_rec[it*bufl][0][0], &buf_write[0], 2*bufl*dNy*dNz, MPI_FLOAT, MPI_SUM, (ncpus-1), MPI_COMM_WORLD);
        if( my_rank == (ncpus-1))  {
             for(ii=0; ii< bufl; ii++)    for(jj=0; jj<dNy; jj++)    for(kk=0; kk<dNz; kk++) {
                 ll=kk+dNz*(jj+ii*dNy);  imag_rec[it*bufl+ii][jj][kk].re=buf_write[ll].re;  imag_rec[it*bufl+ii][jj][kk].im=buf_write[ll].im;   }
        }
   }
   if(res>0) {
        MPI_Reduce(&imag_rec[num*bufl][0][0], &buf_write[0], 2*res*dNy*dNz, MPI_FLOAT, MPI_SUM, (ncpus-1), MPI_COMM_WORLD);
       if( my_rank == (ncpus-1))  {
             for(ii=0; ii< res; ii++)    for(jj=0; jj<dNy; jj++)    for(kk=0; kk<dNz; kk++) {
                 ll=kk+dNz*(jj+ii*dNy);  imag_rec[it*num+ii][jj][kk].re=buf_write[ll].re;  imag_rec[it*num+ii][jj][kk].im=buf_write[ll].im;   }
        }
   }
   DMath::del(buf_write);  

   return 0;
}


#endif  // __PAR__